/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#ifdef CUBIX
#include "extern.h"

static U_DEFINE_STACK (fvar, Fvar*, 10)

int cubix_io_called = 0;

static Fvar *fvar_malloc ()
{
	return (Fvar*) U_malloc (sizeof (Fvar));
}

static char *cubix_fmt_name (lab)
char *lab;
{
	char *str;

	str = (char*) U_malloc (4 + strlen (lab) + 1);
	sprintf (str, "$fmt%s", lab);

	return str;
}

cubix_write_top_sta (sta)
Node *sta;
{
	int type = sta->Statement.Stmt->com.type;

	switch (type) {

	case SUBROUTINE_STMT:
	case FUNCTION_STMT:
	case BLOCK_DATA_STMT:
	case MODULE_STMT:
		write_pass (sta);
		break;

	case IGNORE_STMT:
		break;

	default:
		ostart ();
		otab ();
		oSUBROUTINE ();
		ostr ("CBXMAIN");
		odon ();
		oend ();

		if (type != PROGRAM_STMT)
			write_pass (sta);
	}
}

char *cubix_ioret ()
{
	return "$ioret";
}

/*!!! The following 2 functions are the typical structure used to declare extra
      variables. Note that Etype_INTEGER_0 must be the same in both places and 
      that to prevent bugs we should do it for example getting the etype from the
      entry created by install. This structure is even worse for the opt, auto, ...
      structures done before. Change the strategy.
 */

cubix_decl ()
{
	int i;
	static char *list = "dofioc, dofiol, dolioc, doliol, douioc, douiol, exrdfe, exrdue, exrsfe, exrsle, exrsue, exwdfe, exwdue, exwsfe, exwsle, exwsue, sxrdfe, sxrdue, sxrsfe, sxrsle, sxrsue, sxwdfe, sxwdue, sxwsfe, sxwsle, sxwsue, kclose, kropen, kback, kend, krew, kinqus";

	if (flag_cubix == 0 || !cubix_io_called)
		return;

	otab ();
	ostr (mac_etype_name (Etype_INTEGER_0));
	ostr (list);
	odon ();

	otab ();
	oEXTERNAL ();
	ostr (list);
	odon ();

	otab ();
	ostr (mac_etype_name (Etype_INTEGER_0));
	onam (cubix_ioret ());
	odon ();

	for (i = 0; i < fvar_top; i++) {
		otab ();
		oCHARACTER ();
		ochb ('*');
		oint ((long) fvar [i]->size);
		onam (fvar [i]->var);	
		odon ();
		otab ();
		oPARAMETER ();
		ochb ('(');
		onam (fvar [i]->var);
		ochb ('=');
		ostr (fvar [i]->str);
		ochb (')');
		odon ();
	}
}

cubix_end ()
{
	T_Entry *entry;
	int i;

	if (flag_cubix == 0)
		return;

	entry = install_loc (cubix_ioret (), TAB_NAME, 0);
	entry->typ = typ_malloc ();
	entry->typ->etype = Etype_INTEGER_0;
	SET_temporary (entry);

	for (i = 0; i < fvar_top; i++) {
		entry = install_loc (fvar [i]->var, TAB_NAME, 0);
		entry->typ = typ_malloc ();
		entry->typ->etype = Etype_CHARACTER_0;
		SET_temporary (entry);
	}
}

cubix_end_write ()
{
	cubix_io_called = 0;
	U_stack_clear_fvar ();
}

cubix_checkx_FormatStmt (sta)
Node *sta;
{
	Fvar *fvar = fvar_malloc ();
	char *str, *format;
	int i, size, len, in;

	format = sta->Statement.Stmt->FormatStmt.Format->com.str;

	size = 0;
	len = strlen (format);
	str = (char*) U_malloc (2 * len + 2);

	str [size++] = '\'';
	str [size++] = '(';
	for (i = 0, in = 0; i < len; i++) {
		if (in) {
			if (format [i] == '\'') {
				str [size++] = '\'';
				str [size++] = '\'';
			}
			else
				str [size++] = format [i];
		}	
		else if (format [i] == '(')
			in = 1;
	}
	str [size++] = '\'';
	size = size - 2;

	fvar->label = sta->Statement.IntegerConstant->com.str;
	fvar->size = size;

/*!!! could save some memory by freeing and allocating only the necessary str */

	fvar->str = str;
	fvar->var = cubix_fmt_name (fvar->label);

	U_stack_push_fvar (fvar);
}

Fvar *cubix_get_fvar (label)
char *label;
{
	int i;

	for (i = 0; i < fvar_top; i++)
		if (!strcmp (fvar [i]->label, label))
			return fvar [i];

	SYSERR;

	return fvar [0];
}

cubix_write_Statement (node)
Node *node;
{
/*!!! so that we have the format for write(STRING, ...
	if (node->Statement.Stmt->com.type != FORMAT_STMT)
*/
		write_Statement (node);
}

char *cubix_type (type, size)
int type, size;
{
	if (size == -1) {
		SYSERR;
		return "TYUNKNOWN";
	}

	switch (type) {

	case TYP_INTEGER:
		switch (size) {
		case 2:
			return "2";
		case 4:
			return "3";
		default:
			SYSERR;
			return "0";
		}

	case TYP_REAL:
		switch (size) {
		case 4:
			return "4";
		case 8:
			return "5";
		default:
			SYSERR;
			return "0";
		}

	case TYP_COMPLEX:
		switch (size) {
		case 8:
			return "6";
		case 16:
			return "7";
		default:
			SYSERR;
			return "0";
		}

	case TYP_CHARACTER:
		return "9";
		
	case TYP_LOGICAL:
		return "8";

	default:
		SYSERR;
	}
}

static Node *label;
static char *return_var;
static int add_if, need_extra_if;

static cubix_add_if ()
{
	otab ();
	oIF ();
	oLP ();
	onam (return_var);
	oREL_OP_NE ();
	ochb ('0');
	oRP ();
	oGO_TO ();
	olab (label);
}

cubix_write_IoStmt (node, io_type)
Node *node;
int io_type;
{
	char k1, k2, k3, k4, k5, k6, k7;
	Unit *unit;
	int flagp;
	Fvar *fvar;
	Node *list;

	switch (io_type) {

	case IO_IS_WRITE:
		unit = node->WriteStmt.unit;
		break;

	case IO_IS_READ:
		unit = node->ReadStmt.unit;
		break;

	case IO_IS_PRINT:
		unit = node->PrintStmt.unit;
		break;
	}

	switch (io_type) {

	case IO_IS_WRITE:
	case IO_IS_PRINT:
		k3 = 'w';
		break;

	case IO_IS_READ:
		k3 = 'r';
		break;
	}

	if (unit->rec == NN)
		k4 = 's';
	else
		k4 = 'd';

	if (unit->fmt == NN)
		k5 = 'u';
	else if (unit->fmt == &Times)
		k5 = 'l';
	else
		k5 = 'f';

	flagp = 0;
	if (unit->iostat != NN || unit->err != NN)
		flagp = flagp + 1;
	if (unit->end != NN)
		flagp = flagp + 2;

	if (unit->iostat != NN)
		return_var = unit->iostat->com.str;
	else
		return_var = cubix_ioret ();

	add_if = 0;
	if (unit->iostat != NN || unit->end != NN || unit->err != NN)
		add_if = 1;

	need_extra_if = 0;
	if (unit->iostat != NN || (unit->end != NN && unit->err != NN))
		need_extra_if = 1;

	if (need_extra_if)
		label = label_make ();
	else {
		if (unit->end != NN)
			label = unit->end;
		else if (unit->err != NN)
			label = unit->err;
		else
			label = NN;
	}

#ifdef CHECK
	if ((add_if && label == NN) || (add_if == 0 && label != NN))
		SYSERR;
#endif

/* START */

	otab ();
	onam (return_var);
	ochb ('=');
	ochb ('s'); och ('x'); och (k3); och (k4); och (k5); och ('e');
	ochb ('(');

	if (unit->unit == NN) {
		if (io_type == IO_IS_PRINT)
			ochb ('6');
		else if (io_type == IO_IS_READ)
			ochb ('5');
		else
			SYSERR;
	}
	else if (unit->unit == &Times) {
		switch (io_type) {

		case IO_IS_WRITE:
			ochb ('6');
			break;

		case IO_IS_READ:
			ochb ('5');
			break;

#ifdef CHECK
		default:
			SYSERR;
#endif
		}
	}
	else {
		onod (unit->unit);
	}

	ochb (',');

	if (unit->rec != NN)
		onod (unit->rec);
	else
		ochb ('0');

	ochb (',');

	oint ((long) flagp);

	ochb (',');

	if (unit->fmt == NN || unit->fmt == &Times) {
		ochb ('0');
		ochb (',');
		ostr ("'  '");
	}		
	else if (unit->fmt->com.type == INTEGER_CONSTANT) {
		fvar = cubix_get_fvar (unit->fmt->com.str);
		ochb ('0');
		ochb (',');
		onam (fvar->var);
	}
	else {
		ochb ('0');
		ochb (',');
		onod (unit->fmt);
	}

	ochb (')');
	odon ();

	if (add_if) {
		cubix_add_if ();
		odon ();
	}
	
/* DO */

	switch (io_type) {
	case IO_IS_WRITE:
		list = node->WriteStmt.OutputItemList;
		break;
	case IO_IS_PRINT:
		list = node->PrintStmt.OutputItemList;
		break;
	case IO_IS_READ:
		list = node->ReadStmt.InputItemList;
		break;
	}

	cubix_write_ImpliedValueList (list, k5);

/* END */

	otab ();
	onam (return_var);
	ochb ('=');
	ochb ('e'); och ('x'); och (k3); och (k4); och (k5); och ('e');
	ostr ("()");

	if (add_if && (unit->err != NN || unit->end != NN)) {
		odon ();
		cubix_add_if ();
	}

/* END DONE */

	if (need_extra_if) {
		odon ();
		olab (label);
		otab ();
		oCONTINUE ();
		if (unit->err != NN || unit->end != NN)
			odon ();

		if (unit->err != NN) {
			otab ();
			oIF ();															
			oLP ();
			onam (return_var);
			oREL_OP_GT ();
			ochb ('0');
			oRP ();
			oGO_TO ();
			onod (unit->err);
			if (unit->end != NN)
				odon ();
		}

		if (unit->end != NN) {
			otab ();
			oIF ();															
			oLP ();
			onam (return_var);
			oREL_OP_LT ();
			ochb ('0');
			oRP ();
			oGO_TO ();
			onod (unit->end);
		}
	}
}

cubix_write_ImpliedValueList (list, form)
Node *list;
char form;
{
	Node *ele, *temp;
	Etype *etype;
	int type;
	struct S_ImpliedDo *ido;

	list = list_invert (list);

	for (LIST_LOOP (ele, list, temp)) {
		if (ele->com.type == IMPLIED_DO) {
			ido = &ele->ImpliedDo;

			write_ImpliedDoControl_head (ido->ImpliedDoControl);
			cubix_write_ImpliedValueList (ido->ImpliedValueList, form);
			otab (); oEND_DO (); odon ();
		}
		else {
			otab ();
			onam (return_var);
			ochb ('=');
			ochb ('d'); och ('o'); och (form); och ('i'); och ('o');

			etype = ele->com.typ->etype;
			type = etype->type;
			if (type == TYP_CHARACTER)
				och ('c');
			else
				och ('l');

			ochb ('(');

			ostr (cubix_type (type, mac_etype_size (etype)));

			ochb (',');

			if (ele->com.typ->shape != SN)
				onod (typ_shape_size (ele->com.typ->shape));
			else
				ochb ('1');

			ochb (',');

			oint ((long) mac_etype_size (etype));

			ochb (',');

			if (ele->com.typ->shape != SN) {
				if (ele->com.type != ARRAY_SECTION)
					SYSERR;
				else
					onam (ele->ArraySection.Name->com.str);
			}
			else
				onod (ele);

			ochb (')');
			odon ();

			if (add_if) {
				cubix_add_if ();
				odon ();
			}
		}
	}
}

cubix_write_FilposStmt (node)
Node *node;
{
	struct S_BackspaceStmt *stmt = &node->BackspaceStmt;
	Unit *unit = stmt->unit;

	if (unit->iostat != NN)
		return_var = unit->iostat->com.str;
	else
		return_var = cubix_ioret ();

	otab ();
	onam (return_var);
	oEQ ();

	switch (stmt->Backspace->com.type) {

	case BACKSPACE:
		ostr ("kback");
		break;

	case END_FILE:
		ostr ("kend");
		break;

	case REWIND:
		ostr ("krew");
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	oLP ();

	onod (unit->unit);

	oCOMMA ();

	if (unit->err != NN)
		ochb ('1');
	else
		ochb ('0');

	oRP ();

	if (unit->err != NN) {
		odon ();

		otab ();
		oIF ();
		oLP ();
		onam (return_var);
		oREL_OP_NE ();
		ozero ();
		oRP ();
		oGO_TO ();
		onod (unit->err);
	}
}

cubix_write_StopStmt (node)
Node *node;
{
	char *str;

	otab ();
	oCALL ();
	ostr ("kstop");
	oLP ();

	if (node->StopStmt.StopCode != NN) {
		if (node->StopStmt.StopCode->com.type == CHARACTER_CONSTANT)
			ostr (node->StopStmt.StopCode->com.str);
		else {
			str = U_malloc (strlen (node->StopStmt.StopCode->com.str) + 2 + 1);
			sprintf (str, "'%s'", node->StopStmt.StopCode->com.str);
			ostr (str);
			U_free (str);
		}
	}
	else
		ostr ("'  '");

	oCOMMA ();
	ozero ();
	oRP ();
}

cubix_write_InquireStmt (node)
Node *node;
{
	Unit *unit = node->InquireStmt.unit;

    if (unit->iostat != NN)
        return_var = unit->iostat->com.str;
    else
        return_var = cubix_ioret ();

/* START */

	otab ();
	onam (cubix_ioret ());
	ochb ('=');
	ostr ("kinqus");
	oLP ();

	if (unit->unit != NN)
		onod (unit->unit);
	else
		ostr ("-1");

	oCOMMA ();

	if (unit->file != NN)
		onod (unit->file);
	else
		ostr ("'  '");

	oCOMMA ();

	ozero ();

	oCOMMA ();

	if (unit->err) {
		oone ();
		label = unit->err;
		add_if = 1;
	}
	else {
		ozero ();
		add_if = 0;
	}

	oRP ();

	if (add_if) {
		odon ();
		cubix_add_if ();
	}

/* REST */

	inquire_spec_out (unit->exist, 'l', "1");
	inquire_spec_out (unit->opened, 'l', "2");
	inquire_spec_out (unit->number, 'i', "4");
	inquire_spec_out (unit->named, 'l', "4");
	inquire_spec_out (unit->name, 'c', "1");
	inquire_spec_out (unit->access, 'c', "2");
	inquire_spec_out (unit->sequential, 'c', "4");
	inquire_spec_out (unit->direct, 'c', "8");
	inquire_spec_out (unit->form, 'c', "16");
	inquire_spec_out (unit->formatted, 'c', "32");
	inquire_spec_out (unit->unformatted, 'c', "64");
	inquire_spec_out (unit->recl, 'i', "1");
	inquire_spec_out (unit->nextrec, 'i', "2");
	inquire_spec_out (unit->blank, 'c', "128");
}

static inquire_spec_out (unitspec, ch, index)
Node *unitspec;
char ch, *index;
{
	if (unitspec != NN) {
		odon ();

		otab ();
		onam (return_var);
		oEQ ();
		ostr ("kinqu");
		och (ch);
		oLP ();
		ostr (index);
		oCOMMA ();
		onod (unitspec);
		oRP ();

		if (add_if) {
			odon ();
			cubix_add_if ();
		}
	}
}

cubix_ns ()
{
	ostr ("' '");
}

#else
cubix_decl () {};
cubix_end () {};
cubix_end_write () {};
cubix_checkx_FormatStmt () {};
cubix_write_IoStmt () {};
cubix_io_called () {};
cubix_write_top_sta () {};
cubix_ns () {};
cubix_ioret () {};
cubix_write_StopStmt () {};
cubix_write_FilposStmt () {};
cubix_write_InquireStmt () {};
#endif
